<?php

namespace NoccyLabs\TinyData;

use PDO;
use NoccyLabs\TinyData\Schema\Schema;
use NoccyLabs\TinyData\Schema\TableSchema;
use NoccyLabs\TinyData\Schema\KeyValueSchema;
use NoccyLabs\TinyData\Schema\DocumentSchema;
use NoccyLabs\TinyData\Exception\InvalidSchemaException;

/**
 *
 *
 *
 */
class DataStore
{
    /** @var array The mapped schemas */
    protected $schemas = [];

    /** @var PDO The database */
    protected $db = null;

    /**
     * Constructor, can take a database filename. If none is provided, will
     * default to the script invoked and add a .db extension.
     *
     * @param string|null $dbfile The name of the database file to open/create
     */
    public function __construct($dbfile=null)
    {
        if (!$dbfile) {
            $dbfile = $_SERVER['SCRIPT_NAME'].".db";
        }

        $this->openDatabase($dbfile);
    }

    /**
     * Open database file, called from constructor.
     *
     * @param string $filename The filename
     */
    protected function openDatabase($filename)
    {
        $dbh = new PDO("sqlite:{$filename}", null, null, [
            PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION            
        ]);
        // Create the meta table
        $meta = $dbh->exec(
            "CREATE TABLE IF NOT EXISTS tinydata__meta ".
            "(name TEXT UNIQUE NOT NULL, version INT NOT NULL, definition TEXT NOT NULL);"
        );
        $this->db = $dbh;
    }

    /**
     * Read metadata state for a schema from the database.
     *
     * @param string $name The schema name
     * @return object The schema metadata
     */
    protected function getSchemaMetadata($name)
    {
        $query = $this->db->prepare("SELECT * FROM tinydata__meta WHERE name=:name");
        $query->execute([':name'=>$name]);
        $meta = $query->fetchAll();
        if (count($meta)==0) { 
            return (object)[
                'name' => $name,
                'version' => -1,
                'definition' => []
            ];
        } else {
            $rec = reset($meta);
            return (object)[
                'name' => $rec['name'],
                'version' => intval($rec['version']),
                'definition' => unserialize($rec['definition'])
            ];
        }
    }

    /**
     * Write metadata state for a schema to the database.
     *
     * @param string $name The schema name
     * @param int $version The version of the schema
     * @param array $definition The table definition
     */
    protected function setSchemaMetadata($name, $version, array $definition)
    {
        $query = $this->db->prepare("REPLACE INTO tinydata__meta (name,version,definition) VALUES (:name,:version,:definition);");
        $query->execute([
            ':name' => $name,
            ':version' => $version,
            ':definition' => serialize($definition)
        ]);
    }

    /**
     * Map a schema and make it accessible.
     *
     * @param Schema $schema The schema to map
     */
    public function map(Schema $schema, $name=null)
    {
        // Get the table information
        $name = $name?:$schema->name;
        $current = $this->getSchemaMetadata($name);

        // Create or upgrade the schema if the current version is lower than the
        // provided schema version.
        if ($schema->version > $current->version) {
            $this->upgradeSchema($name, $schema, $current);
        }

        // Create collection for schema
        $collection_type = $schema->getCollectionType();
        $collection = new $collection_type($this, $schema, $name);
        $this->schemas[$name] = $collection;

        return $collection;
    }

    /**
     * Create or upgrade a schema
     * 
     * @param string $name The schema name
     * @param Schema $schema The schema 
     * @param object $current The current schema metadata
     */
    private function upgradeSchema($name, Schema $schema, $current)
    {
        (DEBUG) && printf("Upgrading %s to version %d (was %d)\n", $name, $schema->version, $current->version);
        $actions = [];
        // the database need updating
        $table_def = $schema->definition;
        foreach ($table_def as $col=>$defn) {
            $actions[] = sprintf("%s %s", $col, $defn);
        }

        $stmt = sprintf("CREATE TABLE %s_%d (%s)", $name, $schema->version, join(", ",$actions));
        $this->execute($stmt);

        foreach ($schema->indexes as $index=>$defn) {
            $this->execute("DROP INDEX IF EXISTS {$index}");
            if ($defn->unique) {
                $stmt = sprintf("CREATE UNIQUE INDEX %s ON {$name}_{$schema->version} (%s)", $index, join(",",$defn->columns));
            } else {
                $stmt = sprintf("CREATE INDEX %s ON {$name}_{$schema->version} (%s)", $index, join(",",$defn->columns));
            }
            $this->execute($stmt);
        }

        if ($current->version >= 0) {
            // select old data
            $new_cols = array_keys($table_def);
            $old_cols = array_keys($current->definition);
            $columns = [];
            foreach ($old_cols as $col) {
                if (in_array($col,$new_cols)) $columns[] = $col;
            }
            $columns = join(",",$columns);
            $stmt = "INSERT OR REPLACE INTO {$name}_{$schema->version} ({$columns}) SELECT {$columns} FROM {$name}";
            $this->execute($stmt);
            // then drop old table, 
            $this->execute("DROP TABLE {$name}");
        }
        $this->execute("ALTER TABLE {$name}_{$schema->version} RENAME TO {$name}");

        $this->setSchemaMetadata($name, $schema->version, $schema->definition);

    }

    /**
     * Helper alias for map()
     *
     * @param string $key The schema name
     * @param Schema $value The schema to map
     */
    public function __set($key, $value)
    {
        if (!($value instanceof Schema)) {
            throw new InvalidSchemaException("Only schema instances can be mapped");
        }
        if (array_key_exists($key, $this->schemas)) {
            throw new InvalidSchemaException("Schema names can only be mapped once");
        }
        $this->map($value, $key);
    }

    /**
     * 
     *
     * @param string $key The schema name
     * @return Collection The schema collection
     */
    public function __get($key)
    {
        return $this->schemas[$key];
    }


    /**
     *
     *
     */
    public function execute($sql)
    {
        (DEBUG) && printf("[exec] %s\n", $sql);
        $this->db->exec($sql);
    }

    /**
     *
     *
     */
    public function query($sql, array $fields=[])
    {
        if (DEBUG) printf("[query] %s\n", $sql);
        if (DEBUG) foreach ($fields as $field=>$def) printf("        %s -> %s\n", $field, $def);
        $stmt = $this->db->prepare($sql);
        $stmt->execute($fields);
        $result = $stmt->fetchAll(PDO::FETCH_ASSOC);
        return $result;
    }

    /**
     * Quote a string for a query
     *
     * @param string $str
     * @return string
     */
    public function quote($str)
    {
        return $this->db->quote($str);
    }

    /**
     * Helper to create a new table schema
     *
     * @param string $name The schema name
     * @param int $version The schema version
     * @param array $definition Schema definition
     * @param bool $map If true, the schema will be mapped and the collection be returned
     * @return Schema|Collection The created schema or mapped collection
     */
    public function createTableSchema($name, $version, array $definition=[], $map=false)
    {
        
        $schema = new TableSchema($name, $version, $definition);

        if ($map) {
            return $this->map($schema, $name);
        }
        return $schema;
    }

    /**
     * Helper to create a new key-value schema
     *
     * @param string $name The schema name
     * @param bool $map If true, the schema will be mapped and the collection be returned
     * @return Schema|Collection The created schema or mapped collection
     */
    public function createKeyValueSchema($name, $map=false)
    {
        $schema = new KeyValueSchema($name);

        if ($map) {
            return $this->map($schema, $name);
        }
        return $schema;
    }

    /**
     * Helper to create a new document schema
     *
     * @param string $name The schema name
     * @param int $version The schema version
     * @param array $indexes Columns to index from the documents
     * @param bool $map If true, the schema will be mapped and the collection be returned
     * @return Schema|Collection The created schema or mapped collection
     */
    public function createDocumentSchema($name, $version=1, $indexes=[], $map=false)
    {
        $schema = new DocumentSchema($name, $version, $indexes);

        if ($map) {
            return $this->map($schema, $name);
        }
        return $schema;
    }

}
