<?php

use NoccyLabs\TinyData\DataStore;

class TableTest extends PhpUnit_Framework_TestCase
{
    public function setup()
    {
        $dbf = DBFILE("table");
        $this->ds = new DataStore($dbf);
    }

    public function teardown()
    {
        $this->ds = null;
        $dbf = DBFILE("table");
        if (file_exists($dbf)) unlink($dbf);
    }

    public function testCreatingSchema()
    {
        $testa = $this->ds->createTableSchema("testa", 1);
        $testa
            ->setColumn("foo")
            ->setColumn("bar")
            ->setIndex("foo",["foo","bar"],true);
        $this->ds->testa = $testa;

        $testb = $this->ds->createTableSchema("testb", 1);
        $testb
            ->setColumn("foo")
            ->setColumn("bar")
            ->setIndex("foo",["foo","bar"],true);
        $this->ds->map($testb, "testb");

        $this->ds->createTableSchema("testc", 1, [
            '!name',
            'age' => 'INT'
        ], true);

        $create = exec("sqlite3 ".DBFILE('table')." .dump | grep 'CREATE TABLE \"testa\"'");
        $index = exec("sqlite3 ".DBFILE('table')." .dump | grep 'CREATE UNIQUE INDEX' | grep testa");

        $this->assertEquals("CREATE TABLE \"testa\" (foo TEXT, bar TEXT);", $create);
        $this->assertEquals("CREATE UNIQUE INDEX idx_testa_foo ON \"testa\" (foo,bar);", $index);

        $create = exec("sqlite3 ".DBFILE('table')." .dump | grep 'CREATE TABLE \"testb\"'");
        $index = exec("sqlite3 ".DBFILE('table')." .dump | grep 'CREATE UNIQUE INDEX' | grep testb");

        $this->assertEquals("CREATE TABLE \"testb\" (foo TEXT, bar TEXT);", $create);
        $this->assertEquals("CREATE UNIQUE INDEX idx_testb_foo ON \"testb\" (foo,bar);", $index);

        $create = exec("sqlite3 ".DBFILE('table')." .dump | grep 'CREATE TABLE \"testc\"'");
        $index = exec("sqlite3 ".DBFILE('table')." .dump | grep 'CREATE UNIQUE INDEX' | grep testc");

        $this->assertEquals("CREATE TABLE \"testc\" (name , age INT);", $create);
        $this->assertEquals("CREATE UNIQUE INDEX idx_testc_name ON \"testc\" (name);", $index);
    }

    public function testInsertFind()
    {
        $testa = $this->ds->createTableSchema("testa", 1);
        $testa
            ->setColumn("foo")
            ->setColumn("bar")
            ->setIndex("foo",["foo","bar"],true);
        $this->ds->testa = $testa;

        $this->ds->testa->insert([
            'foo' => 42,
            'bar' => 'heartofgold'
        ]);
        $this->ds->testa->insert([
            'foo' => 3.14,
            'bar' => 'zarniwoop'
        ]);

        $rec = $this->ds->testa->findAll();
        $this->assertEquals(count($rec),2);

        $rec = $this->ds->testa->find([ 'foo'=>42 ]);
        $this->assertEquals(count($rec),1);

        $rec = $this->ds->testa->find([ 'foo gt'=>42 ]);
        $this->assertEquals(count($rec),0);

        $rec = $this->ds->testa->find([ 'foo gte'=>42 ]);
        $this->assertEquals(count($rec),1);

        $rec = $this->ds->testa->find([ 'foo >='=>42 ]);
        $this->assertEquals(count($rec),1);

        $rec = $this->ds->testa->find([ 'foo lt'=>42 ]);
        $this->assertEquals(count($rec),1);

        $rec = $this->ds->testa->find([ 'foo lte'=>42 ]);
        $this->assertEquals(count($rec),2);

        $rec = $this->ds->testa->find([ 'foo <='=>42 ]);
        $this->assertEquals(count($rec),2);

        $rec = $this->ds->testa->findOne([ 'foo'=>42 ]);
        $this->assertEquals(42, $rec['foo']);

    }

    public function testInsertReplace()
    {
        $testa = $this->ds->createTableSchema("testa", 1);
        $testa
            ->setColumn("foo")
            ->setColumn("bar")
            ->setIndex("foo",["foo"],true);
        $this->ds->testa = $testa;

        $this->ds->testa->insert([
            'foo' => 42,
            'bar' => 'heartofgold'
        ]);
        $this->ds->testa->replace([
            'foo' => 42,
            'bar' => 'vogon'
        ]);

        $rec = $this->ds->testa->findAll();
        $this->assertEquals(count($rec),1);

        $rec = $this->ds->testa->findOne([ 'foo'=>42 ]);
        $this->assertEquals("vogon",$rec['bar']);

    }

    public function testInsertUpdate()
    {
        $testa = $this->ds->createTableSchema("testa", 1);
        $testa
            ->setColumn("foo")
            ->setColumn("bar")
            ->setIndex("foo",["foo"],true);
        $this->ds->testa = $testa;

        $this->ds->testa->insert([
            'foo' => 42,
            'bar' => 'heartofgold'
        ]);
        $this->ds->testa->update([
            'bar' => 'vogon'
        ], [
            'foo' => 42,
        ]);

        $rec = $this->ds->testa->findAll();
        $this->assertEquals(count($rec),1);

        $rec = $this->ds->testa->findOne([ 'foo'=>42 ]);
        $this->assertEquals("vogon",$rec['bar']);

    }

    public function testInsertDelete()
    {
        $testa = $this->ds->createTableSchema("testa", 1);
        $testa
            ->setColumn("foo")
            ->setColumn("bar")
            ->setIndex("foo",["foo","bar"],true);
        $this->ds->testa = $testa;

        $this->ds->testa->insert([
            'foo' => 42,
            'bar' => 'heartofgold'
        ]);

        $this->ds->testa->delete(['foo'=>42]);

        $rec = $this->ds->testa->find([ 'foo'=>42 ]);
        $this->assertEquals(count($rec),0);

        $rec = $this->ds->testa->findOne([ 'foo'=>42 ]);
        $this->assertNull($rec);

    }

    public function testInsertDeleteAll()
    {
        $testa = $this->ds->createTableSchema("testa", 1);
        $testa
            ->setColumn("foo")
            ->setColumn("bar")
            ->setIndex("foo",["foo","bar"],true);
        $this->ds->testa = $testa;

        $this->ds->testa->insert([
            'foo' => 42,
            'bar' => 'heartofgold'
        ]);

        $this->ds->testa->deleteAll();

        $rec = $this->ds->testa->find([ 'foo'=>42 ]);
        $this->assertEquals(count($rec),0);

        $rec = $this->ds->testa->findOne([ 'foo'=>42 ]);
        $this->assertNull($rec);

    }

    public function testUpgradingSchema()
    {

        $testa = $this->ds->createTableSchema("testa", 1);
        $testa
            ->setColumn("foo")
            ->setColumn("bar")
            ->setIndex("foo",["foo","bar"],true);
        $this->ds->testa = $testa;

        $this->setup();

        $testa = $this->ds->createTableSchema("testa", 2);
        $testa
            ->setColumn("foo")
            ->setColumn("bar")
            ->setColumn("baz", "INT")
            ->setIndex("foo",["foo","bar"],true);
        $this->ds->testa = $testa;

        $create = exec("sqlite3 ".DBFILE('table')." .dump | grep 'CREATE TABLE \"testa\"'");
        $index = exec("sqlite3 ".DBFILE('table')." .dump | grep 'CREATE UNIQUE INDEX' | grep testa");

        $this->assertEquals("CREATE TABLE \"testa\" (foo TEXT, bar TEXT, baz INT);", $create);
        $this->assertEquals("CREATE UNIQUE INDEX idx_testa_foo ON \"testa\" (foo,bar);", $index);
    }
}
